

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2006 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <list>
#include <time.h>
#include <stdlib.h>
#include <algorithm>

#include "plink.h"
#include "options.h"
#include "helper.h"
#include "crandom.h"
#include "perm.h"
#include "sets.h"
#include "linear.h"
#include "phase.h"


// Global variables
using namespace std;

ofstream LOG;
string PVERSION;
string PDATE;

int main(int argc, char* argv[]) 
{

  /////////////////////////
  // Setup, display title

  CRandom::srand(time(0));
  cout.setf(ios::fixed);
  cout.precision(4);

  set_new_handler(NoMem);

  PVERSION = "0.99n";       // 5 chars
  PDATE    = "09/Oct/2006"; // 11 chars


  //////////////////
  // The major class

  Plink P;


  //////////////////////////
  // Command line arguments

  CArgs a(argc,argv);
  setOptions(a);


  //////////////////////////
  // Start logging, title

  LOG.open(string(par::output_file_name + ".log").c_str());
  P.printLOG("\n"
	     "@----------------------------------------------------------@\n"
	     "|         PLINK!       |    v"+PVERSION+"     |   "+PDATE+"     |\n"
	     "|----------------------------------------------------------|\n"
	     "|  (C) 2006 Shaun Purcell, GNU General Public License, v2  |\n"
	     "|----------------------------------------------------------|\n"
	     "|       http://pngu.mgh.harvard.edu/purcell/plink/         |\n"
	     "@----------------------------------------------------------@\n"
	     "\n");


  /////////////////////
  // Permutation class

  Perm perm(P);


  ////////////////
  // Check version 
  
  if (par::web_check)
    P.webcheck();
  else
    P.printLOG("Skipping web check... [ --noweb ] \n");
  

  /////////////
  // Time stamp

  P.printLOG("Writing this text to log file [ "+
	     par::output_file_name + ".log ]\n");
    
  time_t curr=time(0);
  string tdstamp = ctime(&curr);
  P.printLOG("Analysis started: " + tdstamp +"\n");


  /////////////////////////////////////
  // Validate and record all arguments

  a.check_unused_options(P);

  
  //////////////////////////
  // Some basic definitions

  if (par::species_dog) defineDogChromosomes();
  else defineHumanChromosomes();


  /////////////////////////
  // Input files

  // Binary or ASCII format	  
  if (par::read_bitfile) P.readBinData();
  else P.readData();  

  // Check for duplicate individual or SNP names
  checkDupes(P);

  // Set number of individuals
  P.n = P.sample.size();
  
  // Set number of pairs
  P.np = (int)((double)(P.n*(P.n-1))/(double)2);
  
  // Total number of all (test+background) loci
  P.nl_all = P.locus.size();

  // Number of permutations to store
  P.maxr2.resize(par::replicates);


  /////////////////////////////////////
  // Merge with a secondary data file 
  // Standard (non-list) mode

  if (par::merge_data && !par::merge_list)
    {
      
      if (par::merge_binary)
	P.mergeBinaryData();
      else
	P.mergeData();
      
      // Reset number of individuals
      P.n = P.sample.size();
      
      // Set number of pairs
      P.np = (int)((double)(P.n*(P.n-1))/(double)2);
      
      // Total number of all (test+background) loci
      P.nl_all = P.locus.size();
    }
  
  
  /////////////////////////////////////
  // Merge with a secondary data file 
  // List mode
  
  if (par::merge_list)
    P.mergeList();
      
  

  //////////////////////////////////////////  
  // A different phenotype file specified?

  if (par::pheno_file) P.readPhenoFile();


  //////////////////////////////// 
  // A covariate file specified?

  if (par::covar_file) 
    {
      if (!P.readCovariateFile()) 
	error("Problem reading the specified covariate from the covariate file");
    }
  

  ////////////////////////////////
  // Remove any individuals with 
  // missing phenotypes
  
  if (!par::ignore_phenotypes)
    removeMissingPhenotypes(P);
  

  //////////////////////////////////
  // Binary affection status coding

  if (par::bt)
    affCoding(P);
        
  
  //////////////////////////////////////////////////////////
  // Output a specific set of SNPs (--extract or --exclude)
  
  if (par::dump_set)
    P.extractExcludeSet(false);      
  else if (par::exclude_set)
    P.extractExcludeSet(true);
  
  /////////////////////////////////////////////////////////////
  // Output a specific set of individuals --remove or --keep
    
  if (par::remove_indiv)
    P.removeIndividuals(par::remove_keep_indiv);


  /////////////////////////////////
  // Flip DNA strand for any SNPs? 

  if (par::flip_strand)
    P.flipStrand();

 
 
  //////////////////////////////////////
  // Asign cluster solution from file 

  if (par::include_cluster_from_file)
    {
      P.printLOG("Reading clusters from [ " + 
		 par::include_cluster_filename+" ]\n");
      if (!P.readClusterFile())
	error("Problem reading from [ "+par::include_cluster_filename+" ]");
    }
  

  /////////////////////////////////////////////
  // Assign cluster "solution" to FAM_ID column

  if (par::sol_family)
    {
      
      P.printLOG("Setting to permute within family ID only\n");
      vector<string> famlist;
      P.kname.resize(0);
      for (int i=0; i<P.n; i++)
	{
	  Individual * person = P.sample[i];
	  bool match = false;
	  for (unsigned int j=0; j<famlist.size(); j++)
	    if (person->fid == famlist[j]) 
	      {
  		match=true;
		person->sol=j;
	      }	  
	  if (!match)
	    {
	      famlist.push_back(person->fid);
	      person->sol=famlist.size()-1;
	      P.kname.push_back(person->fid);
	    }
	}

      // Set number of clusters/families
      P.nk = famlist.size();     
    }


  //////////////////////////////////
  // Determine formats for printing
  
  P.prettyPrintLengths();
  

  ///////////////////////////////////////////////
  // Calcalate, read in or fix allele frequencies
  // and performing filtering on missingness, etc
    
    P.printLOG("Before frequency and genotyping pruning, there are "+int2str(P.nl_all)+" SNPs\n");
    
    if (!par::FIXED)
    {
      if (par::SNP_major)
	P.filterSNPs_SNP();
      else
	P.filterSNPs_individual();
    }
  else
    for (int i=0; i<P.nl_all;i++)
      {
	if (P.locus[i]->allele1=="1")
	  P.locus[i]->freq = par::FIX_p;
	else 
	  P.locus[i]->freq = 1-par::FIX_p;
      }

  P.printLOG("After frequency and genotyping pruning, there are "+int2str(P.nl_all)+" SNPs\n");



  ///////////////////////
  // Create family units?
  
  if (par::MENDEL_test || 
      par::MENDEL_report || 
      par::TDT_test ||
      par::QTDT_test &&
      !par::built_families)
    {
      map<string,Individual*> fnd;
      map<Individual*,int> idmap;
      P.linkRelateds(idmap, fnd);
      P.parseTrios();      
      par::built_families = true;
      

      //////////////////////////////////
      // MENDEL error checks (perform now, so 
      // that --make-bed can save an file with 
      // mendel errors zeroed out
      
      if (par::MENDEL_report || par::MENDEL_test)
	P.checkMendel();     

    }

  

  /////////////////////////////////////////////
  // Various output formats and transformations

  if (par::write_bitfile) 
    {
      P.write_BITFILE();
      shutdown();
    }

  if (par::recode) 
    {
      P.display_recoded_PEDFILE();
      shutdown();
    }

  if (par::recode_12) 
    {
      P.display_recoded_PEDFILE_12();
      shutdown();
    }

  if (par::recode_AD) 
    {
      P.display_recoded_PEDFILE_AD();
      shutdown();
    }

  if (par::list_by_allele)
    {
      P.display_listByAllele();
      shutdown();
    }

  if (par::plist)
    {
      P.display_pairList();
    }


  //////////////////////////////////////////
  // Warn if not enough markers in analysis

  if (par::plink || par::cluster || par::genome_output || par::inbreeding)
    {
      if (P.nl_all < 1000) 
	P.printLOG("\n **Warning** this analysis option requires whole-genome level data\n"
		   "             to give accurate results ( probably >10K SNPs ) \n\n");
      }



  ///////////////////////////////////////////////////
  // General class for haplotype phasing and testing
  
  P.haplo = new HaploPhase(P);

  // In case families are included, build family structure if not
  // already done

  if (par::phase_snps || par::mishap_test) 
    {

      // Read in list of tests, or make sliding window?

      if (par::phase_snps) 
	{
	  if (par::sliding_window)
	    P.haplo->makeSlidingWindow(par::sliding_window_size);
	  else
	    P.haplo->readTagFile();
	}
      
      if (par::has_nonfounders && !par::built_families)
	{
	  map<string,Individual*> fnd;
	  map<Individual*,int> idmap;
	  P.linkRelateds(idmap, fnd);
	  P.parseTrios();
	  P.checkMendel();      
	  par::built_families = true;
	}

      P.printLOG("Estimating haplotype frequencies/phases ( MAF >= " 
		 + dbl2str(par::min_af)+" )\n");
      P.printLOG("Considering phases P(H|G) >= " 
		 +dbl2str(par::hap_min_phase_prob)+"\n");
      P.printLOG("Requiring per individual per haplotype genotyping of " 
		 +dbl2str(par::hap_missing_geno)+" \n");

      // Count number of founders
      
      P.haplo->cnt_f = 0;
      vector<Individual*>::iterator person = P.sample.begin();
      while ( person != P.sample.end() ) 
	{  
	  if ( (*person)->founder ) P.haplo->cnt_f++;	    
	  person++;
	}
      if (P.haplo->cnt_f<P.n)
	{
	  P.haplo->nonfounders = true;
	  P.printLOG("Initial phasing based on "+
		     int2str(P.haplo->cnt_f)+" founders ("+
		     int2str(P.n-P.haplo->cnt_f)+
		     " non-founders)\n");
	}
      
      if (P.n == P.haplo->cnt_f && par::test_hap_TDT)
	error("Can not perform TDT in sample with no non-founders");
      
      // Weights (temporarily exclude this from analysis)
      //   ofstream WGT;
      //   if (false)
      //     {
      //       string f = par::output_file_name + ".weight";
      //       P.printLOG("Writing SNP-weights to [ " + f + " ]\n");
      //       WGT.open(f.c_str(), ios::out);
      //     }
      
      //   if (!par::silent) cout << "\n";
      //   if (false)
      //     WGT.close();

    }


  ////////////////////
  // Impute tags SNPs? 
  
  if (par::phase_snps && par::impute_tags)
    {
      
      P.printLOG("Imputing most-likely haplotypes ( P(H|G) >= " + dbl2str(par::hap_post_prob) + " )\n"); 
      
      // Storage space if imputing new SNPs
      P.haplo->new_one.resize(P.n);
      P.haplo->new_two.resize(P.n);
      for (int i=0; i<P.n; i++)
	{
	  P.haplo->new_one[i].resize(0);
	  P.haplo->new_two[i].resize(0);
	}
      

      P.haplo->imputeAllHaplotypes();
      
      shutdown();
    }


  /////////////////////////
  // Haplotype frequencies
  
  if (par::phase_snps && par::display_hap_freqs)
    {
      P.haplo->calculateHaplotypeFrequencies();
      shutdown();
    }

  
  ////////////////////////////////
  // Haplotype phase probabilities
  
  if (par::phase_snps && par::display_phase_probs)
    {
      P.haplo->calculateHaplotypeFrequencies();
      shutdown();
    }
    
  /////////////////////////////////////////////
  // Haplotypic test of non-random missing data
  
  if (par::mishap_test)
    {
      P.performMisHapTests();
      shutdown();
    }
    

  /////////////////////////////////
  // Haplotype association analysis
  
  if (par::phase_snps && ( par::test_hap_CC || par::test_hap_TDT ) )
    {
      
      ////////////////////////////////////////////////
      // Perform omnibus and haplotype-specific tests
      
      string f;
      
      if ( par::test_hap_CC ) f =  par::output_file_name + ".assoc.hap";
      else if ( par::test_hap_TDT ) f = par::output_file_name + ".tdt.hap";
      
      if (par::test_hap_CC)
	{
	  
	  P.printLOG("Writing haplotype association statistics to [ " + f + " ]\n");
	  P.haplo->HTEST.open(f.c_str(), ios::out);
	  P.haplo->HTEST.precision(4);
	  
	  P.haplo->HTEST << setw(10) << "LOCUS" << " " 
			 << setw(12) << "HAPLOTYPE" << " "
			 << setw(10) << "F_A" << " "
			 << setw(10) << "F_U" << " " 
			 << setw(10) << "CHISQ" << " "
			 << setw(4) << "DF" << " "
			 << setw(10) << "P" << " "
			 << "SNPS" << "\n";
	}
      
      
      if (par::test_hap_TDT)
	{
	  P.printLOG("Writing haplotype TDT statistics to [ " + f + " ]\n");
	  P.haplo->HTEST.open(f.c_str(), ios::out);
	  P.haplo->HTEST.precision(4);
	  
	  P.haplo->HTEST << setw(10) << "LOCUS" << " " 
			 << setw(12) << "HAPLOTYPE" << " "
			 << setw(10) << "T" << " "
			 << setw(10) << "U" << " " 
			 << setw(10) << "CHISQ" << " "
			 << setw(10) << "P" << " "
			 << "SNPS" << "\n";
	}
      
      
      P.haplo->phaseAllHaplotypes();

      P.haplo->HTEST.close();
      shutdown();
    }


  /////////////////////////////////
  // Perform a cluster analysis 

  if (par::cluster) 
    { 
      P.buildCluster();
      shutdown();
    }
  


  ////////////////////////////////////////////////////
  // Precalculate frequency-averaged P(IBD|IBS) table

  if (par::plink || par::genome_output) 
    P.preCalcGenomeIBD();

  
  //////////////////////////////
  // Genome-wide output only
 
  if (par::genome_output)
    P.displayGenomeWideInfo();
      
  
  
  //////////////////////////////////////
  // Genome-wide inbreeding output only
 
  if (par::inbreeding)
    {
      
      if (par::SNP_major) 
	P.SNP2Ind();

      ofstream HET;
      string f = par::output_file_name + ".het";
      HET.open(f.c_str(),ios::out);
      HET.precision(4);
        
      P.printLOG("Writing individual heterozygosity information to [ "+f+" ] \n");
      HET << setw(par::pp_maxfid) << "FID" << " "
	  << setw(par::pp_maxiid) << "IID" << " "
	  << setw(12) << "O(HOM)" << " " 
	  << setw(12) << "E(HOM)" << " " 
	  << setw(12) << "N(NM)" << " " 
	  << setw(12) << "F" << "\n";
      
      for (int i1=0; i1<P.n; i1++)
	P.calcInbreeding(P.sample[i1],0,P.nl_all-1,HET);
      
      HET.close();
      
  }



  ///////////////////////
  // Runs of homozygosity
  
  if (par::homo_run)
    {
      
      if (par::SNP_major) P.SNP2Ind();

      ofstream HOM;
      string f = par::output_file_name + ".hom";
      HOM.open(f.c_str(),ios::out);
      HOM.precision(4);
      
      P.printLOG("\nWriting homozygosity-run information to [ "+f+" ] \n");
      string msg = "Run defined as "+int2str(par::homo_run_length);
      if (par::homo_run_length_kb) msg += " kb\n";
      else msg += " SNPs\n";      
      msg += "Allowing "+int2str(par::homo_run_het)+
	" heterozygotes per run (--homo-run-het)\n";
      P.printLOG(msg);
    
      HOM << setw(par::pp_maxfid) << "FID" << " "
	  << setw(par::pp_maxiid) << "IID" << " "
	  << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp) << "SNP1" << " "
	  << setw(par::pp_maxsnp) << "SNP2" << " "
	  << setw(12) << "POS1" << " "
	  << setw(12) << "POS2" << " "
	  << setw(10) << "KB" << " "
	  << setw(10) << "NSNP" << " "
	  << setw(4) << "NHET" << " "
	  << setw(4) << "NMISS" << "\n";


      for (int i1=0; i1<P.n; i1++)
	{
	  if (!par::silent)
	    cout << i1+1 << " of " << P.n << " individuals      \r";
	  P.findHomoRuns(P.sample[i1], HOM);
	}

      if (!par::silent)
	cout << "\n\n";
      
      HOM.close();
  
      // Look for shared homozygous regions
      // if --homo-match option given
      
      if (par::homo_summary_allelic_match)
	P.summariseHomoRuns();
      
      shutdown();
  }
  

  //////////////
  // Runs of IBS
  
  if (par::ibs_run)
    {
      
      if (par::SNP_major) 
	P.SNP2Ind();

      P.segment.resize(0);

      ofstream IBS;
      string f = par::output_file_name + ".ibs";
      if (par::ibs_2only) f += "2";
      IBS.open(f.c_str(),ios::out);
      IBS.precision(4);

      if (par::ibs_2only) 
	P.printLOG("Writing IBS(2)-run information to [ "+f+" ] \n");
      else
	P.printLOG("Writing IBS(1+2)-run information to [ "+f+" ] \n");

      P.printLOG("Run defined as "+int2str(par::ibs_run_length_kb) + " kb ");
      P.printLOG("and "+int2str(par::ibs_run_length_snps) + " SNPs ");
      P.printLOG("(maximum gap of "
		 +int2str( (int)((double)par::ibs_inter_snp_distance/1000))+" kb)\n");
      
      if (par::ibs_2only) 
	P.printLOG("Allowing "+int2str(par::ibs_run_0)+" IBS 0/1 SNPs per run\n");
      else 
	P.printLOG("Allowing "+int2str(par::ibs_run_0)+" IBS 0 SNPs per run\n");

      P.printLOG("Allowing "
		 +int2str(par::ibs_run_missing)+" missing SNPs per run\n");

    
      IBS << setw(par::pp_maxfid) << "FID1" << " "
	  << setw(par::pp_maxiid) << "IID1" << " "
	  << setw(par::pp_maxfid) << "FID2" << " "
	  << setw(par::pp_maxiid) << "IID2" << " "
	  << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp) << "SNP1" << " "
	  << setw(par::pp_maxsnp) << "SNP2" << " "
	  << setw(12) << "POS1" << " "
	  << setw(12) << "POS2" << " "
	  << setw(10) << "KB" << " "
	  << setw(5) << "CM" << " "
	  << setw(5) << "NSNP" << " "
	  << setw(6) << "NIBS0" << " "
	  << setw(6) << "HET" << " "
	  << setw(6) << "NMISS" << " "
	  << setw(6) << "SNPGAP" << " "
	  << setw(6) << "KBGAP" << "\n";
            
      int c=0;
      for (int i1=0; i1<P.n-1; i1++)
	for (int i2=i1+1; i2<P.n; i2++)
	  {
	    if (!par::silent) 
	      cout << ++c << " of " << P.np << " pairs         \r";
	    P.findIBSRuns(P.sample[i1], P.sample[i2], IBS);
	  }
      
      if (!par::silent) 
	cout << "\n\n";
      
      IBS.close();

      // Display summary (i.e. per locus)

      P.printLOG("Found "+int2str(P.segment.size())+" segments\n");
      
      f += ".summary";
      P.printLOG("Writing IBS-run summary to [ "+f+" ]\n\n");
      P.summaryIBSsegments(perm);
      
      shutdown();
  }



  ///////////////////////////////////
  // Runs of missingness (deletions)
  
  if (par::miss_run)
    {

      if (par::SNP_major) P.SNP2Ind();

      ofstream RUN;
      string f = par::output_file_name + ".rum";
      RUN.open(f.c_str(),ios::out);
      
      P.printLOG("Writing run-of-missings information to [ "+f+" ] \n");

      string msg = "Run defined as " + int2str(par::miss_run_length);
      if (par::miss_run_length_kb) msg += " kb\n";
      else msg += " SNPs\n";
      P.printLOG(msg);

      stringstream s2;
      s2 << "With at least " << par::miss_run_level << " missingness\n";
      P.printLOG(s2.str());

      for (int i1=0; i1<P.n; i1++)
	{
	  if (!par::silent)
	    cout << i1+1 << " of " << P.n << " individuals      \r";
	  P.findMissRuns(P.sample[i1],RUN);
	}

      if (!par::silent)
	cout << "\n\n";
      
      RUN.close();

      shutdown();
  }
  
  

  //////////////////////////////
  // Genome-wide association test

  if (par::assoc_test)
    {
      if (par::CMH_test_1 || par::CMH_test_2)
	  P.calcMH();
      else if (par::OR_homog_test)
	  P.calcHomog();
      else if (par::QTDT_test)
	{
	  // Force a Mendel error check
	  if (! (par::MENDEL_report || par::MENDEL_test) )
	    P.checkMendel();
	  P.perm_testQTDT(perm);
	}
      else if (par::boot)
	  P.calcAssociationWithBootstrap();
      else
	  P.calcAssociationWithPermutation(perm);
      shutdown();
    }


  //////////////////////////////
  // QTL interaction test

  if (par::assoc_gxe)
    {
      P.perm_testGXE2(perm);
      shutdown();
    }


  /////////////////////////
  // Hotelling's T^2 test

  if (par::hotel)
    {
      P.perm_testHotel(perm);
      shutdown();
    }


  ///////////////////////////////////
  // Test difference in missing rates

  if (par::test_missing)
    {	
      P.calcAssociationWithPermutation(perm);
      shutdown();
    }
  

  
  
  //////////////////////////////////
  // Genome-wide family-based (TDT)
  // and Parent-of-origin analysis

  if (par::TDT_test)
    {

      // Force a Mendel error check, if we have not 
      // already

      if (! (par::MENDEL_report || par::MENDEL_test) )
	P.checkMendel();

      // Either basic TDT or Parent-Of-Origin analysis

      if (par::parent_of_origin)
	P.perm_testTDT_POO(perm);
      else	  
	P.perm_testTDT(perm);
      shutdown();
    }
  

  //////////////////////////////
  // Genome-wide IBS sharing test

  if (par::ibs_sharing_test)
    {
      P.perm_sharingIBSTest(perm);
      shutdown();
    }


  //////////////////////////////
  // Gene-based test of epistasis

  if (par::epi_genebased)
    {
      P.driverSCREEPI();
      shutdown();
    }


  //////////////////////////////
  // Genome-wide epistasis tests

  if (par::epistasis)
    {
      P.calcEpistasis();      
      shutdown();
    }



  //////////////////////////////
  // LD statistics

  if (par::disp_r1 || par::disp_r2)
    {
      P.calcLDStatistics();
      shutdown();
    }


  ///////////////////
  // LD-based pruning
  
  if (par::prune_ld)
    {  
      P.pruneLD();
      shutdown();
    }

  
 
  

  //////////////////
  // PLINK analysis
  

  // Stop now, unless a plink analysis is specified
  if (!par::plink)
    shutdown();


  if (par::SNP_major) 
    P.SNP2Ind();



  //////////////////////////////////////////////
  // Read pre-computed segment list and perform 
  // segmental tests?

  if (par::read_segment_file)
    {

      ifstream SEG;
      SEG.open(par::read_segment_filename.c_str(),ios::in);
      P.printLOG("Reading IBD-segment information from [ "+par::read_segment_filename+" ]\n");
      checkFileExists(par::read_segment_filename);
      P.readSegmentFile(SEG);
      SEG.close();      
      
      // Perform analysis of segments
      P.summaryIBSsegments(perm);
      shutdown();
    }


  
  //////////////////////////////
  // Pair inclusion/exclusion
  
  // Number of informative pairs
  int c=0; 
  
  // Read or calculate informative pairs?
  if (par::ibd_read) 
    c = P.readInformative();
  else
    c = P.calcInformative();
  

  
  ///////////////////////////////////
  // Save pairs to be included? i.e. 
  // after removing all pairs for 

  // a) low IBD
  // b) not being an affected pair
  // c) being a concordant unaffected pair

  if (par::inc_write) 
    P.writeInformative();

  /////////////////////////////////////////////////////////////////
  // Get and display information on chromosomal range to be tested
  
  else if (par::singlepoint)
    P.printLOG("Using singlepoint analysis mode\n");
  else if (par::inter_grid>0)
    {
      stringstream s2;
      s2 << "Using multipoint analysis: step = " 
	 << par::inter_grid 
	 << " and fringe = " 
	 << par::fringe << " cM\n";
      P.printLOG(s2.str());
    }
  else
    {
      stringstream s2;
      s2 << "Using multipoint analysis: grid = " 
	 << par::grid 
	 << " and fringe = " 
	 << par::fringe << " cM\n";
      P.printLOG(s2.str());
    }

  vector<int> chrs;
  if (par::run_chr==0) {
    vector<int> r = getChromosomeRange(P);
    if (r[0]<1) r[0]=1;
    if (r[1]>22) r[1]=22;
    P.printLOG("\nScanning from chromosome "+
	       int2str(r[0])+" to "+
	       int2str(r[1])+"\n\n");
    for (int i=r[0];i<=r[1];i++) 
      if ( ( !par::chr_haploid[i] ) && 
	   ( !par::chr_sex[i] ) )
	chrs.push_back(i);
  } else chrs.push_back(par::run_chr);
  

  // Blank list of segments
  //  P.segment.resize(0);

  ofstream SEG;
  if (par::segment_output)
    {
      string f = par::output_file_name + ".segment";
      SEG.open(f.c_str(),ios::out);
      P.printLOG("Writing IBD-segment information to [ "+f+" ]\n");
      
      SEG << setw(10) << "FID1" << " "
	  << setw(10) << "IID1" << " "
	  << setw(10) << "FID2" << " "
	  << setw(10) << "IID2" << " ";

      if (par::bt) SEG << setw(4) << "PHE" << " ";
      
      SEG << setw(4) << "CHR" << " "
	  << setw(10) << "BP1" << " "
	  << setw(10) << "BP2" << " " 
	  << setw(par::pp_maxsnp) << "SNP1" << " "
	  << setw(par::pp_maxsnp) << "SNP2" << " "
	  << setw(6) << "NSNP" << " "
	  << setw(10) << "KB" << "\n";

      f = par::output_file_name + ".segment.summary";
      P.printLOG("Writing IBD-segment summary to [ "+f+" ]\n\n");
      
      P.printLOG("Segment thresholds are "+dbl2str(par::segment_threshold_start)
		 +" and "+dbl2str(par::segment_threshold_finish)+"\n");
      P.printLOG("Maximum intra-segment inter-SNP distance is "
		 +int2str(par::segment_inter_snp_distance)
		 +"\n");


    }


  ofstream MP;
  if (par::multi_output)
    {
      string f = par::output_file_name + ".multi";
      MP.open(f.c_str(), ios::out);
      MP.precision(5);
      P.printLOG("Writing multipoint IBD estimates to [ "+ f+" ]\n");
    }

  ofstream GMULTI;
  if (par::gmulti_output)
    {
      string f = par::output_file_name + ".gmulti";
      GMULTI.open(f.c_str(), ios::out);
      GMULTI.precision(4);
      P.printLOG("Writing genotype/multipoint IBD estimates to [ "+ f +" ]\n");
    }



  ///////////////////////////
  // Consider each chromosome

  for (int i=0;i<chrs.size();i++)
    {
      
      // Reset main variables
      P.phenotype.resize(0);
      P.pihat.resize(0);
      P.Zlocus.resize(0);
      P.m_pihat.resize(0);
      P.v_pihat.resize(0);
      P.pair1.resize(0);
      P.pair2.resize(0);
      
      // Set chromosome
      par::run_chr = chrs[i];
      
      // Find scan range
      P.setMarkerRange();
      
      // Total number of all (test+background) loci
      P.nl_all = P.locus.size();
      
      // Number of loci for test loci
      P.nl = par::run_end - par::run_start + 1; 
      P.printLOG(int2str(P.nl)+" markers in this scan\n");
      
      // Set up single/multipoint marker map
      if (!par::singlepoint)
	P.preCalcMultiPoint();
      else
	P.preCalcSinglePoint();

           
      /////////////////////////////////
      //  For each pair of individuals
      //  calculate mutlipoint pihats
      
      
      // reset counter
      int c1=0;
      int c2=0;
      
      for (int i1=0; i1<P.n-1; i1++)
	for (int i2=i1+1; i2<P.n; i2++)
	  {
	    
	    if (!P.skip_pair[c1++])
	      {
		Individual * p1 = P.sample[i1];
		Individual * p2 = P.sample[i2];
		
		/////////////////////////////////
		//  1. Calculate IBD(g) | IBS(g) 
		
		// Note: could be outside of chromosome loop, 
		// but this way there is no need to store the values
		
		Z IBDg = P.saved_IBDg[c2];
		
		if (!par::silent)
		  {
		    cout << "IBD calculation: " 
			 << ++c2
			 << " of " 
			 << c 
			 << "                  \r";
		    cout.flush();
		  }
		
		/////////////////////////////////
		//  2. Calculate IBD(l) - IBD(g)
		
		vector<Z> IBDl = P.calcLocusIBD(p1,p2,IBDg);
		
		
		/////////////////////////////////
		//  3. Multipoint calculation
				
		P.pairid = itoa((int)p1->phenotype,10)
		  +" "+itoa((int)p2->phenotype,10)+" ";
		P.pairid += itoa((int)c2,10) + " ";
		P.pairid += p1->fid+"_"+p1->iid+"_ ";
		P.pairid += p2->fid+"_"+p2->iid+"_";
		
		vector<double> p;
		
		//////////////////////////////////
		// Perform either using
		//     Singlepoint analysis
		//     Multipoint analysis (default)
		
		if (!par::singlepoint) p = P.calcMultiPoint(IBDl,IBDg,MP);
		else p = P.calcSinglePoint(IBDl,IBDg);
		

		////////////////////////////////
		//   3b. Verbose output:
		//       genotypes for each pair
		
		if (par::gmulti_output)
		  {
		    for (int l=par::run_start; l<=par::run_end; l++)
		      P.displayGMULTI(p1,p2,l,GMULTI);			
		  }
		
				

		///////////////////////////////
		//  4. Scan for segments of IBD
		
		bool inseg = false;
		
		Segment s;
		s.p1 = p1;
		s.p2 = p2;
		int npos = p.size();

		for (int l=0; l<npos; l++)
		  {
		    
		    ///////////////////////////
		    // Start of a new segment?

		    if ( (!inseg) && p[l] >= par::segment_threshold_start)
		      {
			inseg = true;
			
			if (P.m1[l] != -1)
			  s.start = P.m1[l];
			else
			  s.start = par::run_start;
		      }
		    
		    
		    ///////////////////////////
		    // End of existing segment?
		    		    
		    if ( inseg && 
			 ( p[l] < par::segment_threshold_finish || l == npos-1 ) )
		      {
			inseg = false;
			
			if (P.m2[l-1] != -1)
			  s.finish = P.m2[l-1];
			else
			  s.finish = par::run_end;

			// Add segment to list
			P.segment.push_back(s);

			// Display?
			if (par::segment_output)
			  {
			    SEG << setw(10) << s.p1->fid << " "
				<< setw(10) << s.p1->iid << " "
				<< setw(10) << s.p2->fid << " "
				<< setw(10) << s.p2->iid << " ";

			    if (par::bt)
			      {
				if ( (!p1->aff) && (!p2->aff) ) 
				  SEG << setw(4) << "-1" << " ";
				else if ( p1->aff && p2->aff )
				  SEG << setw(4) << "1" << " ";
				else if ((!p1->aff) && p2->aff)
				  SEG << setw(4) << "0" << " ";
				else if (p1->aff && !p2->aff)
				  SEG << setw(4) << "0" << " ";
				else
				  SEG << setw(4) << "NA" << " ";
			      }
			    else
			      SEG << setw(4) << "NA" << " ";

			    Locus * start = P.locus[s.start];
			    Locus * finish = P.locus[s.finish];
			    
			    SEG << setw(4) << par::run_chr << " "
				<< setw(10) << start->bp << " "
				<< setw(10) << finish->bp << " "
				<< setw(par::pp_maxsnp) << start->name << " "
				<< setw(par::pp_maxsnp) << finish->name << " "
				<< setw(6) << s.finish-s.start+1 << " "
				<< setw(10) << (double)(finish->bp - start->bp)/(double)1000 
				<< "\n";
			    
			  }
			
		      }
		    		    
		  }
		

		/////////////////////////////
		//  5. Add to list 

		// Do not bother saving for now...
		// only save segments...
		if (false) P.pihat.push_back(p);
		
		// And (A,B) pair to list
		P.pair1.push_back(i1);
		P.pair2.push_back(i2);
		
	      }
  	    	    
	  }
      
      if (!par::silent)
	cout << "\n";
      
      
      /////////////////////////////////
      //  Make list of unique individuals
      
      // copy first set of individuals
      P.in_anal = P.pair1;
      for (unsigned int ind=0; ind<P.pair2.size(); ind++) 
	P.in_anal.push_back(P.pair2[ind]);
      sort(P.in_anal.begin(),P.in_anal.end());
      vector<int>::iterator new_end=
	unique(P.in_anal.begin(), P.in_anal.end());
      // delete all elements past new_end 
      P.in_anal.erase(new_end, P.in_anal.end());
      
	P.printLOG(int2str(P.in_anal.size())+
		   " unique, informative individuals in analysis\n");
      
      if ( P.in_anal.size() == 0 )
	{
	  error("No individuals left in analysis: halting");
	}
      
      /////////////////////////////////
	//  Verbose output: summarise IBD
      
      if (par::segment_output)
	{
	  //	    P.summaryIBDsegments(perm);
	}
      else
	if (par::summary_ibd_output)
	  P.summaryIBD();
      
	
	/////////////////////////////////
	  //  Regression analysis
	  
// 	  P.printLOG("Analysis of "+int2str(P.pair1.size()) +
// 		 " informative pairs at " +int2str(P.pihat[0].size())+
// 		 " positions\n");
      
//       if (par::permute)
// 	P.calcRegression(i);
	  
	  
	  /////////////////////////////////
	    //  Next chromosome
	    
	    par::done_global_pihat = true;
	    
	    if (!par::silent)
	      cout << "\n";
	    
    }
  
  // Now do IBD segment (as IBS...)
  if (par::segment_output)
    P.summaryIBSsegments(perm);
  
  // Shut segment and multipoint files
  if (par::segment_output)
    SEG.close();
  if (par::multi_output)
    MP.close();
  if (par::gmulti_output)
    GMULTI.close();
  
  
  ////////////////////////////////
  //  Output genome-wide p-values

  if (par::permute)
    if (chrs.size()>=1 && (!par::ignore_phenotypes))
      P.displayGenomePV();
  
    

  ////////////////////////////////
  //  Done
  
  shutdown();

}


